/* =============================================================================
 UPC - EETAC - CSD - http://digsys.upc.edu
 P11: An example on how to use the LCD library of functions. 
 P12: Example on how to use the TIMER2 and the TIMER3 

 * Bicycle speed sensor.
Features for this Phase #1 of the project:   
 * Adjusted for a 20" tyre (https://en.wikipedia.org/wiki/Bicycle_wheel)
 * Magnetic sensor mounted on the wheel fork and magnet on a wheel spoke
 * Maximum speed 100 km/h 
 * Target chip PIC18F4520. 
 * Project complexity: P8. Dedicated processor.FSM + datapath (Timer & counters)
 * C language, coding in FSM style, a dedicated processor approach. 
 * The use of the microcontroller TMR2 peripheral to generate a time base.
 * The use of the TMR3 to count external pulses from the bike sensor
 * Start/stop button (INT0)
 * LCD display to represent the bike speed
 * Prototype for design step #5: PICDEM2+ board
 * ON_LED to indicate that the instrument is running. 
 
Approach: 
- Step 1: try ot organise the project as a dedicated processor in P8. 
    - Principle of operation.
    - Datapath components. 
    - FSM design. State diagram. 

- Step 2: Prepare the hardware, so that adapt the PICDEM2+ board to this 
particular application. For instance:
    - ON_LED connected to RB3
    - Speed sensor is the T13CKI at RC0
    - Start/stop is RB0/INT0
    - A switch can be connected to select speed or distance (odometer. (Phase#2)		
 
- Step 3: Program the init_system() function, configuring ports, interrupts 
and the TMR2 and TMR3. 

For instance, in order to generate a timing period of 5.745 s ===> 
FCLK/4 = 1MHz; PRESCALER = 1/4; TMR2 = 250; POSTSCALER = 1/15;
A software variable TB_N4_max = 383 (char) will set the var_TB_flag
    
   5.745 s = 1 us * 4 * 250 * 15 * 383          
                  
This is a TMR2 interrupt every 15 ms, which is going to be used to increment 
the var_TMR2_N4 variable until the limit at Time_Base_N4 is reached 


NOTE:The C code exemplified here is not optimised in any sense. The CSD style of 
programming aims clarity and facilitating the comprehension of the basic ideas 
on microcontrollers taking advantage of the previous content on combinational
and sequential systems.  
============================================================================= */

#include <xc.h>	// This file also includes the "pic18f4520.h" which contains 
                //all the microcontroller declarations
#include <stdio.h>
//#include <math.h>
#include "lcd.h"

/* 	============================================================================
        Definitions
        ===================================================================== */
// This structure allows to access to a given bit into an 8-bit register
#define 	PORTBIT(adr, bit) ( (unsigned)(&adr)*8 + (bit) )

/* This is the state enumeration. Check the state diagram posted on the web to
figure out which is the meaning of every state */
#define 	Idle                'A'
#define 	Setup_Time_Base     'B'
#define 	Timing              'C'
#define 	Meas_speed          'D'
#define 	Stop_Time_Base      'E'

// Max value for the time base TMR2 software postscaler 
#define 	TMR2_N4_max  383 
#define 	TMR2_N2_max  250


/* 	============================================================================
        Function prototypes
    ========================================================================= */
static void interrupt service_routine(void); 
// This is the interrupt service routine

void init_system(void);
char output_logic(void);
char state_logic(void);

void read_inputs(void);  // to interface the hardware pins
void write_outputs(void);


/* 	============================================================================
        Global variables
============================================================================= */

static char var_ON_LED;
static char var_Sel_Speed; // to decide which calculations to perform (Phase #2)
static int var_Speed; // Speed measurement 
static char var_Speed_ASCII[16]; // Measurement translated to ASCII for printing
static char var_current_state = Idle; 	// state variable

static char var_STSP_Flag;    	// This flag is going to be set by interrupt 
								//when clicking the trigger push button
static char var_LCD_Flag;   // This flag is for writing only to the LCD 
						//display only when new information is available.

static char var_TC_Flag;  // This flag is set once the time base has generated the 
                     // gate period for allowing the counting of external pulses
                    //with the TMR1

//The "software" TMR2 post scaler variable to implement a larger counter
static int var_TMR2_N4; //software post_scaler_value for the time base


/* 	============================================================================
        Main function
        ======================================================================*/
void main(void){
    init_system ();
    while(1) {                      /* loop forever the FSM */
	if (state_logic()) break;
	if (output_logic()) break;
    }
    while (1){
        __nop();// The system has crashed in an illegal state and needs 
                //attention. This loop is never reached in normal operation. 
                //The only way to scape it is clicking the CD (Master Reset)
    }   
}

/* 	============================================================================
        Function definitions
        ===================================================================== */

//Future_state_logic routine 
//******************************************************************************
char state_logic(void) {
    char error = 0;
    read_inputs (); // Sample the variables that are going to decide where to go 

    switch (var_current_state) {
        case Idle:
            if (var_STSP_Flag == 1) {   // if ST/SP button is detected,
                var_current_state = Setup_Time_Base;
                var_STSP_Flag = 0;      //Clear variable, 
                                        //so that it can be interrupted again
                var_LCD_Flag = 0;
            } else
                var_current_state = Idle;
            break;

        case Setup_Time_Base:
            var_current_state = Timing;  // In this state the time base is 
                                        // reinitialised 
            break;

        case Timing:
            if (var_TC_Flag == 1){ // End of the timing period
                var_current_state = Meas_speed;
                var_LCD_Flag = 1;   //Write the next message (speed measurement)
            } else
                var_current_state = Timing;
            break;

        case Meas_speed:
            if (var_STSP_Flag == 1) {   // if ST/SP button is detected,
                var_current_state = Stop_Time_Base;
                var_STSP_Flag = 0;      //Clear variable, 
                                        //so that it can be interrupted again
                var_LCD_Flag = 0;
            } else
                var_current_state = Setup_Time_Base; 
                            // to perform another measurement cycle
            break;
            
        case Stop_Time_Base:
            var_current_state = Idle;
            var_STSP_Flag = 0; 
                //Clear variable, so that it can be interrupted again to start
            
            break;

        default:
            error = 1;
    }
    return (error);
}

//Output logic routine 
//******************************************************************************
char output_logic(void) {
    char error = 0;
    switch (var_current_state) {
        case Idle:
            var_ON_LED = 0;
            if (var_LCD_Flag == 1) {    //write LCD only if new information
                lcd_clear();		// Delete 
                lcd_home();			// select line 1
                       //"----------------"  // 16 characters ASCII  
                lcd_puts("     IDLE       ");
                lcd_home2();			// select line 2 to write the result
                lcd_puts(". Push Start   .");
            }
            var_LCD_Flag = 0;

            break;

        case Setup_Time_Base:

            var_ON_LED = 1; // Start operation
            
			// Starting the Timer2 and enable its interrupts 

			var_TMR2_N4	= 0;		// Init the software timer extension (N4))
			PR2 = TMR2_N2_max;
			TMR2ON = 1; 			// Start counting pulses
			TMR2IE = 1;             //Enable interrupts
            var_TC_Flag = 0;        // Init the gate time
            TMR3ON = 1; // TMR3 enable 
            TMR3 = 0; // Clear counting
            break;

        case Timing:
            var_ON_LED = 1; // System operating and counting pulses using TMR1
            
            break;
           
        case Meas_speed:  
            var_ON_LED = 1; // Perform an speed calculation
            TMR2ON = 0; 	// Stop Timer2
			TMR2IE = 0;	
            // Reading number of pulses counted
            var_Speed = TMR3; 
            sprintf(var_Speed_ASCII, "Value = %u", var_Speed);
            
            if (var_LCD_Flag == 1) {    //write LCD only if new information
                lcd_clear();		
                lcd_home();		
                lcd_puts("  Speed (km/h)  ");
                lcd_home2();		
                lcd_puts(var_Speed_ASCII);
            }            
            
            
            var_LCD_Flag = 0;   //Write only one time per measurement
            break;
            
        case Stop_Time_Base:
            var_ON_LED = 0; // Ends the timing period

//            Timer0_Post_Scaler = 0;
//            TMR0IE = 0; // Disable TMR0 interrupts
//            TMR0ON = 0; // Stop Timer0

            var_LCD_Flag = 1; // to print initial LCD information
            break;

        default:
            error = 1;
    }
    write_outputs();// Write the output pins once the output variables are set
    return (error);
}

// Interrupt Service Routine to attend the pushbutton and the Timer0
//******************************************************************************
static void interrupt service_routine(void) { //Interrupt is the key word here
    GIE = 0; //Disable interrupts while attending one of them ...
    if (TMR2IF == 1) { // If TMR2 terminal count ?
        if (var_TMR2_N4 < TMR2_N4_max) {
            var_TMR2_N4++;
        } else {                // The gate time has ended 
            var_TMR2_N4 = 0;
            var_TC_Flag = 1;    // So that the system can measure speed
            TMR3ON = 0;         // TMR3 disables now to save the pulse counting
        } 

        TMR2IF = 0; 
    // Clear the interrupt flag, so that it will be interrupted again in 15 ms
    }
    if (INT0IF == 1) {
        var_STSP_Flag = 1; //Set the variable to move the state diagram and clear 
        // the hardware interrupt flag bit (so that the system can be 
        // interrupted again when another active edge is detected)
        INT0IF = 0;
    }

    GIE = 1; 	//Enable interrupts
}

/* =============================================================================
 Scan inputs and process the bits so that we can set our convenient variables. 
 Many of the variables are "char" because we can easily monitor them using
 the watch window for debugging purposes.
  =========================================================================== */
void read_inputs (void){     
// Like in P9, read the port, mask the bit of interest and set the variable 
    var_Sel_Speed = (PORTA & 0b00010000)>>4;
   // This variable is not used in the Phase #1 of the project (speed only)
}

/* =============================================================================
	Driving the output pins. Let's drive the output ports, converting 
	microcontroller variables into electrical signals.    
   ========================================================================== */
void write_outputs(void){
	char var_buffer; //Local variable buffer
        // keep all the bits except the bit to be written and write all the port
        var_buffer = PORTB & 0b11110111;         
        PORTB = var_buffer | (var_ON_LED << 3);

// Here is where the messages to the LCD should be written depending on the flag 
// and the state         
        
    }


//*****************************************************************************
//Init operations  - System setup 
//*****************************************************************************

void init_system(void) {
// Microcontroller general configuration bits. No need to change them.
// PIC18F4520 Configuration Bit Settings. They can be generated automatically 
// in the MPLAB X IDE

// CONFIG1H
#pragma config OSC = XT       // Oscillator Section bits (XT oscillator)
#pragma config FCMEN = OFF     
// Fail-Safe Clock Monitor E bit (Fail-Safe Clock Monitor disabled)
#pragma config IESO = OFF       
// Internal/External Osc Switch-over bit (Oscillator Switch-over mode disabled)

// CONFIG2L
#pragma config PWRT = ON       // Power-up Timer E bit (PWRT enabled)
#pragma config BOREN = SBORDIS 
// Brown-out Reset E bits (Brown-out Reset enabled in hardware 
// only (SBOREN is disabled))
#pragma config BORV = 3    // Brown Out Reset Voltage bits (Minimum setting)

// CONFIG2H
#pragma config WDT = OFF
// Watchdog Timer E bit (WDT disabled (control is on the SWDTEN bit))
#pragma config WDTPS = 32768    
// Watchdog Timer Postscale Sect bits (1:32768)

// CONFIG3H
#pragma config CCP2MX = PORTC  
// CCP2 MUX bit (CCP2 input/output is multiplexed with RC1)
#pragma config PBADEN = OFF    
// PORTB A/D E bit (PORTB<4:0> pins are configured 
// as digital I/O on Reset)
#pragma config LPT1OSC = OFF   
// Low-Power Timer1 Oscillator E bit (Timer1 configured for 
// higher power operation)
#pragma config MCLRE = ON
// MCLR Pin E bit (MCLR pin enabled; RE3 input pin disabled)
 
// CONFIG4L
#pragma config STVREN = ON      
// Stack Full/Underflow Reset E bit (Stack full/underflow will cause Reset)
#pragma config LVP = OFF
// Single-Supply ICSP E bit (Single-Supply ICSP disabled)
#pragma config XINST = OFF
// Extended Instruction Set E bit (Instruction set extension and Indexed 
// Addressing mode disabled (Legacy mode))

// CONFIG5L
#pragma config CP0 = OFF
// Code Protection bit (Block 0 (000800-001FFFh) not code-protected)
#pragma config CP1 = OFF        
// Code Protection bit (Block 1 (002000-003FFFh) not code-protected)
#pragma config CP2 = OFF        
// Code Protection bit (Block 2 (004000-005FFFh) not code-protected)
#pragma config CP3 = OFF        
// Code Protection bit (Block 3 (006000-007FFFh) not code-protected)

// CONFIG5H
#pragma config CPB = OFF        
// Boot Block Code Protection bit 
// (Boot block (000000-0007FFh) not code-protected)
#pragma config CPD = OFF        
// Data EEPROM Code Protection bit (Data EEPROM not code-protected)

// CONFIG6L
#pragma config WRT0 = OFF       
// Write Protection bit (Block 0 (000800-001FFFh) not write-protected)
#pragma config WRT1 = OFF       
// Write Protection bit (Block 1 (002000-003FFFh) not write-protected)
#pragma config WRT2 = OFF       
// Write Protection bit (Block 2 (004000-005FFFh) not write-protected)

#pragma config WRT3 = OFF       
// Write Protection bit (Block 3 (006000-007FFFh) not write-protected)

// CONFIG6H
#pragma config WRTC = OFF       
// Configuration Register Write Protection bit (Configuration registers 
// (300000-3000FFh) not write-protected)
#pragma config WRTB = OFF
// Boot Block Write Protection bit (Boot block (000000-0007FFh) 
// not write-protected)
#pragma config WRTD = OFF
// Data EEPROM Write Protection bit (Data EEPROM not write-protected)

// CONFIG7L
#pragma config EBTR0 = OFF      
// Table Read Protection bit (Block 0 (000800-001FFFh) not protected from 
// table reads executed in other blocks)
#pragma config EBTR1 = OFF      
// Table Read Protection bit (Block 1 (002000-003FFFh) not protected from 
// table reads executed in other blocks)
#pragma config EBTR2 = OFF      
// Table Read Protection bit (Block 2 (004000-005FFFh) not protected from table 
// reads executed in other blocks)
#pragma config EBTR3 = OFF      
// Table Read Protection bit (Block 3 (006000-007FFFh) not protected from table 
// reads executed in other blocks)

// CONFIG7H
#pragma config EBTRB = OFF      
// Boot Block Table Read Protection bit (Boot block (000000-0007FFh) 
// not protected from table reads executed in other blocks)

    /* -+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+*/
	
// Initialise starts here:
    // By default, let's declare all the unused pins as digital outputs 
// The ports:
    PORTA = 0x00;   // Reset all Flip-Flops at PORTA
    LATA = 0x00;    // Reset all Flip-Flops at PORTA
    ADCON1 = 0x0F;  // Configures all the PORTA and PORTB(3..0) lines as digital
    TRISA = 0b00010000; // RA4 is the Speed/Odometer input
    
    PORTB = 0x00; // Reset all Flip-Flops at PORTB
    LATB = 0x00; // Reset all Flip-Flops at PORTB
    TRISB = 0b00000001; // PORTB RB3 is the ON_LED
                        // PORTB RB0 is the INT0 Start/Stop input
    
    PORTC = 0x00; // Reset all Flip-Flops at PORTC
    LATC = 0x00; // Reset all Flip-Flops at PORTC
    TRISC = 0b00000001; // PORTB RC0 is the Sensor input T13CKI


// The TMR2 configuration to run as a time base for the gate time
    TMR2ON = 0;     // Stopped now
    
	T2CKPS1 = 0;	// Prescaler is 4  ---> N1
	T2CKPS0	= 1;	

	TOUTPS3 = 1;    // Postscaler is 15  --->  N2
	TOUTPS2	= 1;
	TOUTPS1	= 1;
	TOUTPS0	= 0;	
    
    PR2 = 250;  // This is the value to which the TMR2 will be compared ---> N3
    
	TMR2IE 	= 0; // Timer2 interrupts not allowed when initiating the uC
    // Activating the Timer2 interrupts will be done in "Setup_Time_Base" state
	PEIE 	= 1; // Enable interrupts from peripherals (TMR2 in this example)

// The TMR3 for counting external pulses from the sensor
    RD163 = 1; //16 bt reading mode
    T3CKPS1 = 0; // 00 = 1:1 Prescale value
    T3CKPS0 = 0;
    T1OSCEN = 0; // External oscillator disabled. 
    TMR3CS = 1; // External pulses from T13CKI
    TMR3ON = 0; // TMR3 disables now
    TMR3 = 0; // Clear TMR3 
    
    
// The INT0
    INTEDG0 = 0;    // INT0 edge selection: '0' --> on falling edge 
                    //(interrupt after clicking the pushbutton Start/Stop)
    INT0IE = 1;     // Enable interrupts from the INT0 pin

// General interrupt enable:
    GIE = 1;        // Global interrupts allowed

//Initial situation at "Idle" state:
    var_ON_LED = 0;
    var_STSP_Flag = 0;
    var_TC_Flag = 0;

// LCD initialisation 
	lcd_init(FOURBIT_MODE); 	/* Initialise the LCD module  */
	var_LCD_Flag = 1;   // Let's the system print the initial message

}

